/*
 * Copyright(c) 2006 to 2018 ADLINK Technology Limited and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
 */
#ifndef DDSC_DDS_H
#define DDSC_DDS_H

/* All header files distributed with Cyclone DDS reside in dds. This file is
   merely a proxy for dds/dds.h and must be considered deprecated. */
#include "dds/dds.h"

#endif /* DDSC_DDS_H */

